"""
This module contains the definition of several neural networks used in state-of-the-art
federated learning papers.
"""
import string
import sys
from abc import abstractmethod

import torch
import torch.nn as nn
from torch.functional import F
from torchvision.models import resnet18, resnet34, resnet50

sys.path.append(".")
sys.path.append("..")

from . import FlukeENV  # NOQA
from .utils.model import batch_norm_to_group_norm  # NOQA

__all__ = [
    'EncoderHeadNet',
    'GlobalLocalNet',
    'HeadGlobalEncoderLocalNet',
    'EncoderGlobalHeadLocalNet',
    'MNIST_2NN',
    'MNIST_2NN_E',
    'MNIST_2NN_D',
    'MNIST_CNN',
    'MNIST_CNN_E',
    'MNIST_CNN_D',
    'FedBN_CNN',
    'FedBN_CNN_E',
    'FedBN_CNN_D',
    'MNIST_LR',
    'CifarConv2',
    'CifarConv2_E',
    'CifarConv2_D',
    'ResNet9',
    'ResNet9_E',
    'ResNet9_D',
    'FEMNIST_CNN',
    'FEMNIST_CNN_E',
    'FEMNIST_CNN_D',
    'VGG9_E',
    'VGG9_D',
    'VGG9',
    'ResNet18',
    'ResNet34',
    'ResNet50',
    'ResNet18GN',
    'FedAVGCNN_E',
    'FedAVGCNN_D',
    'FedAVGCNN',
    'MoonCNN_E',
    'MoonCNN_D',
    'MoonCNN',
    'LeNet5_E',
    'LeNet5_D',
    'LeNet5',
    'Shakespeare_LSTM_E',
    'Shakespeare_LSTM_D',
    'Shakespeare_LSTM'
]


class EncoderHeadNet(nn.Module):
    r"""Encoder (aka backbone) + Head Network [Base Class]
    This type of networks are defined as two subnetworks, where one is meant to be the
    encoder/backbone network that learns a latent representation of the input, and the head network
    that is the classifier part of the model. The forward method should work as usual (i.e.,
    :math:`g(f(\mathbf{x}))` where :math:`\mathbf{x}` is the input, :math:`f` is the encoder and
    :math:`g` is the head), but the ``forward_encoder`` and ``forward_head`` methods should be used
    to get the output of the encoder and head subnetworks, respectively.
    If this is not possible, they fallback to the forward method (default behavior).

    Args:
        encoder (nn.Module): Encoder subnetwork.
        head (nn.Module): Head subnetwork.
    """

    def __init__(self, encoder: nn.Module, head: nn.Module):
        super(EncoderHeadNet, self).__init__()
        self._encoder = encoder
        self._head = head

    @property
    def encoder(self) -> nn.Module:
        """Return the encoder subnetwork.

        Returns:
            nn.Module: Encoder subnetwork.
        """
        return self._encoder

    @property
    def head(self) -> nn.Module:
        """Return the head subnetwork.

        Returns:
            nn.Module: head subnetwork.
        """
        return self._head

    def forward_encoder(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through the encoder subnetwork.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor of the encoder subnetwork.
        """
        return self._encoder(x)

    def forward_head(self, z: torch.Tensor) -> torch.Tensor:
        """Forward pass through the head subnetwork. ``z`` is assumed to be the output of the
        encoder subnetwork or an "equivalent" tensor.

        Args:
            z (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor of the head subnetwork.
        """
        return self._head(z)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._head(self._encoder(x))


class GlobalLocalNet(nn.Module):
    """Global-Local Network (Abstract Class). This is a network that has two subnetworks, one is
    meant to be shared (global) and one is meant to be personalized (local). The :meth:`forward`
    method should work as expected, but the :meth:`forward_local`` and :meth:`forward_global`
    methods should be used to get the output of the local and global subnetworks, respectively.
    If this is not possible, they fallback to the forward method (default behavior).
    """

    @abstractmethod
    def get_local(self) -> nn.Module:
        """Return the local subnetwork.

        Returns:
            nn.Module: The local subnetwork
        """
        raise NotImplementedError

    @abstractmethod
    def get_global(self) -> nn.Module:
        """Return the global subnetwork.

        Returns:
            nn.Module: The global subnetwork
        """
        raise NotImplementedError

    def forward_local(self, x) -> torch.Tensor:
        return self.get_local()(x)

    def forward_global(self, x) -> torch.Tensor:
        return self.get_global()(x)


class EncoderGlobalHeadLocalNet(GlobalLocalNet):
    """This implementation of the Global-Local Network (:class:`GlobalLocalNet`) is meant to be used
    with the Encoder-Head architecture. The global (i.e., that is shared between clients and server)
    subnetwork is the encoder and the local (i.e., not shared) subnetwork is the head.

    Args:
        model (EncoderHeadNet): The federated model to use.

    See Also:
        - :class:`EncoderHeadNet`
        - :class:`GlobalLocalNet`
        - :class:`HeadGlobalEncoderLocalNet`
    """

    def __init__(self, model: EncoderHeadNet):
        assert isinstance(model, EncoderHeadNet), "model must be an EncoderHeadNet."
        super(EncoderGlobalHeadLocalNet, self).__init__()
        self.model = model

    def get_local(self) -> nn.Module:
        return self.model.head

    def get_global(self) -> nn.Module:
        return self.model.encoder

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class HeadGlobalEncoderLocalNet(GlobalLocalNet):
    """This implementation of the Global-Local Network (:class:`GlobalLocalNet`) is meant to be used
    with the Encoder-Head architecture. The global (i.e., that is shared between clients and server)
    subnetwork is the head and the local (i.e., not shared) subnetwork is the encoder.

    Args:
        model (EncoderHeadNet): The federated model to use.

    See Also:
        - :class:`EncoderHeadNet`
        - :class:`GlobalLocalNet`
        - :class:`EncoderGlobalHeadLocalNet`
    """

    def __init__(self, model: EncoderHeadNet):
        assert isinstance(model, EncoderHeadNet), "model must be an EncoderHeadNet."
        super(HeadGlobalEncoderLocalNet, self).__init__()
        self.model = model

    def get_local(self) -> nn.Module:
        return self.model.encoder

    def get_global(self) -> nn.Module:
        return self.model.head

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class MNIST_2NN_E(nn.Module):
    """Encoder for the :class:`MNIST_2NN` network.

    Args:
        hidden_size (tuple[int, int], optional): Size of the hidden layers. Defaults to (200, 200).

    See Also:
        - :class:`MNIST_2NN`
        - :class:`MNIST_2NN_D`
    """

    def __init__(self,
                 hidden_size: tuple[int, int] = (200, 100)):
        super(MNIST_2NN_E, self).__init__()
        self.fc1 = nn.Linear(28*28, hidden_size[0])
        self.fc2 = nn.Linear(hidden_size[0], hidden_size[1])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.view(-1, 784)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return x


class MNIST_2NN_D(nn.Module):
    """Head for the :class:`MNIST_2NN` network.

    Args:
        hidden_size (int, optional): Size of the hidden layer. Defaults to 200.
        use_softmax (bool, optional): If True, the output is passed through a softmax layer,
            otherwise, no activation will be used. Defaults to False.

    See Also:
        - :class:`MNIST_2NN`
        - :class:`MNIST_2NN_E`
    """

    def __init__(self,
                 hidden_size: int = 100,
                 use_softmax: bool = False):
        super(MNIST_2NN_D, self).__init__()
        self.use_softmax = use_softmax
        self.fc3 = nn.Linear(hidden_size, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.use_softmax:
            return F.softmax(self.fc3(x), dim=1)
        else:
            # return torch.sigmoid(self.fc3(x))
            return self.fc3(x)


# FedAvg: https://arxiv.org/pdf/1602.05629.pdf - hidden_size=[200,200], w/o softmax
# SuPerFed - https://arxiv.org/pdf/2109.07628v3.pdf - hidden_size=[200,200], w/o softmax
# pFedMe: https://arxiv.org/pdf/2006.08848.pdf - hidden_size=[100,100], w/ softmax
# FedDyn: https://openreview.net/pdf?id=B7v4QMR6Z9w - hidden_size=[200,100], w/o softmax
class MNIST_2NN(EncoderHeadNet):
    """Multi-layer Perceptron for MNIST. This is a
    2-layer neural network for MNIST classification first introduced in the [FedAvg]_ paper,
    where the hidden layers have 200 neurons each and the output layer with sigmoid activation.

    Similar architectures are also used in other papers:

    - [SuPerFed]_: ``hidden_size=(200, 200)``, same as FedAvg;
    - [pFedMe]_: ``hidden_size=(100, 100)`` with softmax on the output layer;
    - [FedDyn]_: ``hidden_size=(200, 100)``.

    Args:
        hidden_size (tuple[int, int], optional): Size of the hidden layers. Defaults to (200, 200).
        softmax (bool, optional): If True, the output is passed through a softmax layer.
          Defaults to True.

    See Also:
        - :class:`MNIST_2NN_E`
        - :class:`MNIST_2NN_D`

    References:
        .. [FedAvg] H. Brendan McMahan, Eider Moore, Daniel Ramage, Seth Hampson, Blaise Aguera y
            Arcas. "Communication-Efficient Learning of Deep Networks from Decentralized Data".
            In AISTATS (2017).
        .. [SuPerFed] Seok-Ju Hahn, Minwoo Jeong, and Junghye Lee. Connecting Low-Loss Subspace for
            Personalized Federated Learning. In KDD (2022).
        .. [pFedMe] Canh T. Dinh, Nguyen H. Tran, and Tuan Dung Nguyen. Personalized Federated
            Learning with Moreau Envelopes. In NeurIPS (2020).
        .. [FedDyn] S. Wang, T. Liu, and M. Hong. "FedDyn: A Dynamic Federated Learning Framework".
            In ICLR (2021).
    """

    def __init__(self,
                 hidden_size: tuple[int, int] = (200, 100),
                 softmax: bool = False):
        super(MNIST_2NN, self).__init__(
            MNIST_2NN_E(hidden_size),
            MNIST_2NN_D(hidden_size[1], softmax)
        )


class MNIST_CNN_E(nn.Module):
    """Encoder for the :class:`MNIST_CNN` network.

    See Also:
        - :class:`MNIST_CNN`
        - :class:`MNIST_CNN_D`
    """

    def __init__(self):
        super(MNIST_CNN_E, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 5, 1)
        self.conv2 = nn.Conv2d(32, 64, 5, 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        return x.view(-1, 1024)


class MNIST_CNN_D(nn.Module):
    """Head for the :class:`MNIST_CNN` network.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`MNIST_CNN`
        - :class:`MNIST_CNN_E`
    """

    def __init__(self, output_size: int = 10):
        super(MNIST_CNN_D, self).__init__()
        self.fc1 = nn.Linear(1024, 512)
        self.fc2 = nn.Linear(512, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.fc1(x))
        return self.fc2(x)


# FedAvg: https://arxiv.org/pdf/1602.05629.pdf
# SuPerFed - https://arxiv.org/pdf/2109.07628v3.pdf
# works with 1 channel input - MNIST4D
class MNIST_CNN(EncoderHeadNet):
    """Convolutional Neural Network for MNIST. This is a simple CNN for MNIST classification
    first introduced in the [FedAvg]_ paper, where the architecture consists of two convolutional
    layers with 32 and 64 filters, respectively, followed by two fully connected layers with 512
    and 10 neurons, respectively.

    Very same architecture is also used in the [SuPerFed]_ paper.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.
    """

    def __init__(self, output_size: int = 10):
        super(MNIST_CNN, self).__init__(MNIST_CNN_E(), MNIST_CNN_D(output_size))


class FedBN_CNN_E(nn.Module):
    """Encoder for the :class:`FedBN_CNN` network.

    Args:
        channels (int, optional): Number of input channels. Defaults to 1.

    See Also:
        - :class:`FedBN_CNN`
        - :class:`FedBN_CNN_D`
    """

    def __init__(self, channels: int = 1):
        super(FedBN_CNN_E, self).__init__()

        self.conv1 = nn.Conv2d(channels, 64, 5, 1, 2)
        self.bn1 = nn.BatchNorm2d(64)
        self.conv2 = nn.Conv2d(64, 64, 5, 1, 2)
        self.bn2 = nn.BatchNorm2d(64)
        self.conv3 = nn.Conv2d(64, 128, 5, 1, 2)
        self.bn3 = nn.BatchNorm2d(128)
        # this layer is erroneously reported in the paper
        self.conv4 = nn.Conv2d(128, 128, 5, 1, 2)
        self.bn4 = nn.BatchNorm2d(128)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.bn3(self.conv3(x)))
        x = F.relu(self.bn4(self.conv4(x)))
        return x.view(-1, 6272)


class FedBN_CNN_D(nn.Module):
    """Head for the :class:`FedBN_CNN` network.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`FedBN_CNN`
        - :class:`FedBN_CNN_E`
    """

    def __init__(self, output_size: int = 10):
        super(FedBN_CNN_D, self).__init__()
        self.fc1 = nn.Linear(6272, 2048)
        self.bn5 = nn.BatchNorm1d(2048)
        self.fc2 = nn.Linear(2048, 512)
        self.bn6 = nn.BatchNorm1d(512)
        self.fc3 = nn.Linear(512, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.bn5(self.fc1(x)))
        x = F.relu(self.bn6(self.fc2(x)))
        return self.fc3(x)


# FedBN: https://openreview.net/pdf?id=6YEQUn0QICG
class FedBN_CNN(EncoderHeadNet):
    """Convolutional Neural Network with Batch Normalization for CIFAR-10. This network
    follows the architecture proposed in the [FedBN]_ paper, where the encoder consists of four
    convolutional layers with 64, 64, 128, and 128 filters, respectively, and the head network
    consists of three fully connected layers with 2048, 512, and 10 neurons, respectively.

    Args:
        channels (int, optional): Number of input channels. Defaults to 1.
        output_size (int, optional): Number of output classes. Defaults to 10.

    Note:
        In the original paper, the size of the last convolutional layer is erroneously reported.

    See Also:
        - :class:`FedBN_CNN_E`
        - :class:`FedBN_CNN_D`

    References:
        .. [FedBN] Xiaoxiao Li, Meirui JIANG, Xiaofei Zhang, Michael Kamp, and Qi Dou. FedBN:
            Federated Learning on Non-IID Features via Local Batch Normalization. In ICLR (2021).
    """

    def __init__(self, channels: int = 1, output_size: int = 10):
        super(FedBN_CNN, self).__init__(FedBN_CNN_E(channels), FedBN_CNN_D(output_size))


# FedNH: https://arxiv.org/abs/2212.02758 (CIFAR-10)
class CifarConv2_E(nn.Module):
    """Encoder for the :class:`CifarConv2` network.

    See Also:
        - :class:`CifarConv2`
        - :class:`CifarConv2_D`
    """

    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=64, kernel_size=5)
        self.conv2 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=5)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

    def forward(self, x: torch.Tensor):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 64 * 5 * 5)
        return x


class CifarConv2_D(nn.Module):
    """Head for the :class:`CifarConv2` network.

    See Also:
        - :class:`CifarConv2`
        - :class:`CifarConv2_E`
    """

    def __init__(self):
        super().__init__()
        self.linear1 = nn.Linear(1600, 384)
        self.linear2 = nn.Linear(384, 192)
        self.linear3 = nn.Linear(192, 10, bias=False)

    def forward(self, x: torch.Tensor):
        x = F.relu(self.linear1(x))
        x = F.relu(self.linear2(x))
        logits = self.linear3(x)
        return logits


class CifarConv2(EncoderHeadNet):
    """Convolutional Neural Network for CIFAR-10. This is a CNN for CIFAR-10 classification
    as described in the [FedNH]_ paper, where the architecture consists of two convolutional
    layers with 64 filters, followed by two fully connected layers with 384 and 100 neurons,
    respectively. The convolutional layers are followed by ReLU activations and max pooling.
    The last classification layer is a linear layer with 10 neurons.

    Args:
        embedding_size (int, optional): Size of the embedding after the second linear layer.
            Defaults to 100.
        num_classes (int, optional): Number of classes. Defaults to 10.

    See Also:
        - :class:`CifarConv2_E`
        - :class:`CifarConv2_D`

    References:
        .. [FedNH] Yutong Dai, Zeyuan Chen, Junnan Li, Shelby Heinecke, Lichao Sun, Ran Xu.
            Tackling Data Heterogeneity in Federated Learning with Class Prototypes.
            In AAAI (2023).
    """

    def __init__(self):
        super().__init__(CifarConv2_E(), CifarConv2_D())


# FedProx: https://openreview.net/pdf?id=SkgwE5Ss3N (MNIST and FEMNIST)
# Logistic Regression
class MNIST_LR(nn.Module):
    """Logistic Regression for MNIST. This is a simple logistic regression model for MNIST
    classification used in the [FedProx]_ paper for both MNIST and FEMNIST datasets.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    References:
        .. [FedProx] Tian Li, Anit Kumar Sahu, Manzil Zaheer, Maziar Sanjabi, Ameet Talwalkar, and
            Virginia Smith. Federated Optimization in Heterogeneous Networks. Adaptive & Multitask
            Learning Workshop. In Open Review https://openreview.net/pdf?id=SkgwE5Ss3N (2018).
    """

    def __init__(self, output_size: int = 10):
        super(MNIST_LR, self).__init__()
        self.fc = nn.Linear(784, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.view(-1, 784)
        return F.softmax(self.fc(x), dim=1)


class _ResidualBlock(nn.Module):

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 padding: int,
                 stride: int):
        super(_ResidualBlock, self).__init__()
        self.conv_res1 = nn.Conv2d(in_channels=in_channels,
                                   out_channels=out_channels,
                                   kernel_size=kernel_size,
                                   padding=padding,
                                   stride=stride,
                                   bias=False)
        self.conv_res1_bn = nn.BatchNorm2d(num_features=out_channels, momentum=0.9)
        self.conv_res2 = nn.Conv2d(in_channels=out_channels,
                                   out_channels=out_channels,
                                   kernel_size=kernel_size,
                                   padding=padding,
                                   bias=False)
        self.conv_res2_bn = nn.BatchNorm2d(num_features=out_channels, momentum=0.9)

        if stride != 1:
            # in case stride is not set to 1, we need to downsample the residual so that
            # the dimensions are the same when we add them together
            self.downsample = nn.Sequential(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(num_features=out_channels, momentum=0.9)
            )
        else:
            self.downsample = None

        self.relu = nn.ReLU(inplace=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        out = self.relu(self.conv_res1_bn(self.conv_res1(x)))
        out = self.conv_res2_bn(self.conv_res2(out))

        if self.downsample is not None:
            residual = self.downsample(residual)

        out = self.relu(out)
        out = out + residual
        return out


class ResNet9_E(nn.Module):
    """Encoder for the :class:`ResNet9` network.

    See Also:
        - :class:`ResNet9`
        - :class:`ResNet9_D`
    """

    def __init__(self):
        super(ResNet9_E, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels=3, out_channels=64, kernel_size=3,
                      stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=64, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=64, out_channels=128,
                      kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=128, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            _ResidualBlock(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            nn.Conv2d(in_channels=128, out_channels=256,
                      kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=256, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(in_channels=256, out_channels=256,
                      kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=256, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            _ResidualBlock(in_channels=256, out_channels=256, kernel_size=3, stride=1, padding=1),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.conv(x)
        out = out.view(-1, out.shape[1] * out.shape[2] * out.shape[3])
        return out


class ResNet9_D(nn.Module):
    """Head for the :class:`ResNet9` network.

    See Also:
        - :class:`ResNet9`
        - :class:`ResNet9_E`
    """

    def __init__(self, output_size: int = 100):
        super(ResNet9_D, self).__init__()
        self.fc = nn.Linear(in_features=1024, out_features=output_size, bias=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.fc(x)
        return out


# SuPerFed - https://arxiv.org/pdf/2001.01523.pdf (CIFAR-100)
class ResNet9(EncoderHeadNet):
    """ResNet-9 network for CIFAR-100 classification. This network follows the architecture proposed
    in the [SuPerFed]_ paper, which fllows the standard ResNet-9. The encoder consists of all the
    layers but the last fully connected layer, and thus the head network consists of the last fully
    connected layer.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 100.

    See Also:
        - :class:`ResNet9_E`
        - :class:`ResNet9_D`
    """

    def __init__(self, output_size: int = 100):
        super(ResNet9, self).__init__(ResNet9_E(), ResNet9_D(output_size))


# DITTO: https://arxiv.org/pdf/2012.04221.pdf (FEMNIST)
class FEMNIST_CNN_E(nn.Module):
    """Encoder for the :class:`FEMNIST_CNN` network.

    See Also:
        - :class:`FEMNIST_CNN`
        - :class:`FEMNIST_CNN_D`
    """

    def __init__(self):
        super(FEMNIST_CNN_E, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, kernel_size=5, padding=2)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=5, padding=2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, kernel_size=2, stride=2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, kernel_size=2, stride=2)
        return x.view(-1, 7 * 7 * 64)


class FEMNIST_CNN_D(nn.Module):
    """Head for the :class:`FEMNIST_CNN` network.

    See Also:
        - :class:`FEMNIST_CNN`
        - :class:`FEMNIST_CNN_E`
    """

    def __init__(self):
        super(FEMNIST_CNN_D, self).__init__()
        self.fc1 = nn.Linear(7 * 7 * 64, 1024)
        self.fc2 = nn.Linear(1024, 62)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class FEMNIST_CNN(EncoderHeadNet):
    """Convolutional Neural Network for FEMNIST. This is a simple CNN for FEMNIST classification
    first introduced in the [DITTO]_ paper, where the architecture consists of two convolutional
    layers with 32 and 64 filters, respectively, followed by two fully connected layers with 1024
    and 62 neurons, respectively. Each convolutional layer is followed by a ReLU activation and a
    max pooling layer.

    References:
        .. [DITTO] Tian Li, Shengyuan Hu, Ahmad Beirami, and Virginia Smith. Ditto: Fair and Robust
            Federated Learning Through Personalization. In ICML (2021).
    """

    def __init__(self):
        super(FEMNIST_CNN, self).__init__(FEMNIST_CNN_E(), FEMNIST_CNN_D())


class VGG9_E(nn.Module):
    """Encoder for the :class:`VGG9` network.

    Args:
        seed (int, optional): Seed used for weight initialization. Defaults to 98765.

    See Also:
        - :class:`VGG9`
        - :class:`VGG9_D`
    """

    @classmethod
    def _conv_layer(self,
                    in_channels: int,
                    out_channels: int,
                    kernel_size: int,
                    stride: int = 1,
                    padding: int = 0,
                    groups: int = 1,
                    bias: bool = False,
                    seed: int = 0) -> nn.Conv2d:
        conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size,
                         padding=padding, groups=groups, stride=stride, bias=bias)
        torch.manual_seed(seed)
        torch.nn.init.xavier_normal_(conv.weight)
        return conv

    def __init__(self, seed: int = 98765):
        super(VGG9_E, self).__init__()
        self.encoder = nn.Sequential(
            VGG9_E._conv_layer(in_channels=1, out_channels=16, kernel_size=3,
                               padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            VGG9_E._conv_layer(in_channels=16, out_channels=32, kernel_size=3,
                               padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            VGG9_E._conv_layer(in_channels=32, out_channels=64, kernel_size=3,
                               padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            VGG9_E._conv_layer(in_channels=64, out_channels=128, kernel_size=3,
                               padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            VGG9_E._conv_layer(in_channels=128, out_channels=256,
                               kernel_size=3, padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            VGG9_E._conv_layer(in_channels=256, out_channels=512,
                               kernel_size=3, padding=1, bias=False, seed=seed),
            nn.ReLU(True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Flatten()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.encoder(x)


class VGG9_D(nn.Module):
    """Head for the :class:`VGG9` network.

    Args:
        input_size (int, optional): Size of the input tensor. Defaults to 512.
        output_size (int, optional): Number of output classes. Defaults to 62.
        seed (int, optional): Seed used for weight initialization. Defaults to 98765.

    See Also:
        - :class:`VGG9`
        - :class:`VGG9_E`
    """
    @classmethod
    def _linear_layer(cls,
                      in_features: int,
                      out_features: int,
                      bias: bool = False,
                      seed: int = 0) -> nn.Linear:
        fc = nn.Linear(in_features, out_features, bias=bias)
        torch.manual_seed(seed)
        torch.nn.init.xavier_normal_(fc.weight)
        return fc

    def __init__(self, input_size: int = 512, output_size: int = 62, seed: int = 98765):
        super(VGG9_D, self).__init__()
        self.downstream = nn.Sequential(
            VGG9_D._linear_layer(in_features=input_size, out_features=256, bias=False, seed=seed),
            nn.ReLU(True),
            VGG9_D._linear_layer(in_features=256, out_features=output_size, bias=False, seed=seed)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.downstream(x)


# SuPerFed: https://arxiv.org/pdf/2109.07628v3.pdf (FEMNIST)
class VGG9(EncoderHeadNet):
    """VGG-9 network for FEMNIST classification. This network follows the architecture proposed in
    the [SuPerFed]_ paper which follows the standard VGG-9 architecture. In this implementation
    all convolutional layers are considered as the encoder and the fully connected layers are
    considered as the head network.

    Args:
        input_size (int, optional): Size of the input tensor. Defaults to 784.
        output_size (int, optional): Number of output classes. Defaults to 62.
        seed (int, optional): Seed used for weight initialization. Defaults to 98765.

    See Also:
        - :class:`VGG9_E`
        - :class:`VGG9_D`
    """

    def __init__(self, output_size: int = 62, seed: int = 98765):
        super(VGG9, self).__init__(
            VGG9_E(seed),
            VGG9_D(input_size=512, output_size=output_size, seed=seed)
        )


# FedAvg: https://arxiv.org/pdf/1602.05629.pdf (CIFAR-10)
# FedDyn: https://openreview.net/pdf?id=B7v4QMR6Z9w (CIFAR-10 and CIFAR-100)
class FedAVGCNN_E(nn.Module):
    """Encoder for the :class:`FedAVGCNN` network.

    See Also:
        - :class:`FedAVGCNN_D`
        - :class:`FedAVGCNN`
    """

    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=5, stride=1, padding=2)
        self.pool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.conv2 = nn.Conv2d(64, 64, kernel_size=5, stride=1, padding=2)
        self.pool2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.conv1(x))
        x = self.pool1(x)
        x = F.relu(self.conv2(x))
        x = self.pool2(x)
        x = x.view(x.size(0), -1)
        return x


class FedAVGCNN_D(nn.Module):
    """Head for the :class:`FedAVGCNN` network.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`FedAVGCNN_E`
        - :class:`FedAVGCNN`
    """

    def __init__(self, output_size: int = 10):
        super().__init__()
        self.local3 = nn.Linear(4096, 384)
        self.local4 = nn.Linear(384, 192)
        self.linear = nn.Linear(192, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.local3(x))
        x = F.relu(self.local4(x))
        x = self.linear(x)
        return F.softmax(x, dim=1)


class FedAVGCNN(EncoderHeadNet):
    """Convolutional Neural Network for CIFAR-10.
    This network follows the architecture proposed in the  FedAVG paper as well as in FedDyn.
    The encoder consists of two convolutional layers with 64 filters, followed by two fully
    connected layers with 384 and 192 neurons, respectively. The convolutional layers are followed
    by ReLU activations and max pooling. The last classification layer is a linear layer.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`FedAVGCNN_E`
        - :class:`FedAVGCNN_D`
    """

    def __init__(self, output_size: int = 10):
        super(FedAVGCNN, self).__init__(FedAVGCNN_E(),
                                        FedAVGCNN_D(output_size))


# FedOpt: https://openreview.net/pdf?id=SkgwE5Ss3N (CIFAR-10)
class ResNet18(nn.Module):
    """ResNet-18 network as defined in the torchvision library.

    Note:
        This class is a wrapper around the ResNet-18 model from torchvision and it does not
        implement the :class:`EncoderHeadNet` interface.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.
    """

    def __init__(self, output_size=10):
        super(ResNet18, self).__init__()
        self.resnet = resnet18(num_classes=output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.resnet(x)


class ResNet18GN(ResNet18):
    """ResNet-18 network as defined in the torchvision library but with Group Normalization layers
    instead of Batch Normalization.

    Note:
        This class is a wrapper around the ResNet-18 model from torchvision and it does not
        implement the :class:`EncoderHeadNet` interface.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.
    """

    def __init__(self, output_size: int = 10):
        super(ResNet18GN, self).__init__(output_size)
        batch_norm_to_group_norm(self)


# FedPer: https://arxiv.org/pdf/1912.00818.pdf (CIFAR-100)
class ResNet34(nn.Module):
    """ResNet-34 network as defined in the torchvision library.

    Note:
        This class is a wrapper around the ResNet-18 model from torchvision and it does not
        implement the :class:`EncoderHeadNet` interface.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 100.
    """

    def __init__(self, output_size: int = 100):
        super(ResNet34, self).__init__()
        self.resnet = resnet34(num_classes=output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.resnet(x)


# MOON: https://arxiv.org/pdf/2103.16257.pdf (CIFAR-100)
class ResNet50(nn.Module):
    """ResNet-50 network as defined in the torchvision library.

    Note:
        This class is a wrapper around the ResNet-18 model from torchvision and it does not
        implement the :class:`EncoderHeadNet` interface.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 100.
    """

    def __init__(self, output_size: int = 100):
        super(ResNet50, self).__init__()
        self.resnet = resnet50(num_classes=output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.resnet(x)


class LeNet5_E(nn.Module):
    """Encoder for the :class:`LeNet5` network.

    See Also:
        - :class:`LeNet5`
        - :class:`LeNet5_D`
    """
    # Expected input size: 32x32x3

    def __init__(self):
        super(LeNet5_E, self).__init__()
        self.layer1 = nn.Sequential(
            nn.Conv2d(3, 6, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(6),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2))
        self.layer2 = nn.Sequential(
            nn.Conv2d(6, 16, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.layer1(x)
        out = self.layer2(out)
        out = out.reshape(out.size(0), -1)
        return out


class LeNet5_D(nn.Module):
    """Head for the :class:`LeNet5` network.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 100.

    See Also:
        - :class:`LeNet5`
        - :class:`LeNet5_E`
    """

    def __init__(self, output_size: int = 10):
        super(LeNet5_D, self).__init__()
        self.fc = nn.Linear(400, 120)
        self.relu = nn.ReLU()
        self.fc1 = nn.Linear(120, 84)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Linear(84, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.fc(x)
        out = self.relu(out)
        out = self.fc1(out)
        out = self.relu1(out)
        out = self.fc2(out)
        return out


# FedRep: https://arxiv.org/pdf/2102.07078.pdf (CIFAR-100 and CIFAR-10)
# LG-FedAvg: https://arxiv.org/pdf/2001.01523.pdf
class LeNet5(EncoderHeadNet):
    """LeNet-5 for CIFAR. This is a LeNet-5 for CIFAR-10/100 classification as described in the
    [FedRep]_ paper, where the architecture consists of two convolutional layers with 6 and 16
    filters, respectively, followed by two fully connected layers with 120 and 84 neurons,
    respectively, and the output layer with 10 neurons. The activation functions are ReLU. The
    architecture is also used in the [LG-FedAvg]_ paper.

    See Also:
        - :class:`LeNet5_E`
        - :class:`LeNet5_D`

    Args:
        output_size (int, optional): Number of output classes. Defaults to 100.

    References:
        .. [FedRep] Liam Collins, Hamed Hassani, Aryan Mokhtari, and Sanjay Shakkottai.
            Exploiting shared representations for personalized federated learning. In ICML (2021).
        .. [LG-FedAvg] Paul Pu Liang, Terrance Liu, Liu Ziyin, Nicholas B. Allen, Randy P. Auerbach,
            David Brent, Ruslan Salakhutdinov, Louis-Philippe Morency. Think Locally, Act Globally:
            Federated Learning with Local and Global Representations.
            In arXiv https://arxiv.org/abs/2001.01523 (2020).
    """

    def __init__(self, output_size: int = 10):
        super(LeNet5, self).__init__(LeNet5_E(), LeNet5_D(output_size))


# SuPerFed: https://arxiv.org/pdf/2109.07628v3.pdf (Shakespeare)
class Shakespeare_LSTM_E(nn.Module):
    """Encoder for the :class:`Shakespeare_LSTM` network.

    See Also:
        - :class:`Shakespeare_LSTM`
        - :class:`Shakespeare_LSTM_D`
    """

    def __init__(self):
        super(Shakespeare_LSTM_E, self).__init__()
        self.encoder = nn.Embedding(256, 8)
        self.rnn = nn.LSTM(
            input_size=8,
            hidden_size=256,
            num_layers=2,
            batch_first=True,
            bias=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x, _ = self.rnn(x)
        return x[:, -1, :]


class Shakespeare_LSTM_D(nn.Module):
    """Head for the :class:`Shakespeare_LSTM` network.

    See Also:
        - :class:`Shakespeare_LSTM`
        - :class:`Shakespeare_LSTM_E`
    """

    def __init__(self):
        super(Shakespeare_LSTM_D, self).__init__()
        self._output_size = len(string.printable)
        self.classifier = VGG9_D._linear_layer(
            256, self._output_size, bias=False, seed=FlukeENV().get_seed())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.classifier(x)


class Shakespeare_LSTM(EncoderHeadNet):
    """LSTM for Shakespeare. This is an LSTM for Shakespeare classification first introduced
    in the [SuPerFed]_ paper, where the architecture consists of an embedding layer with 8
    dimensions, followed by a two-layer LSTM with 256 hidden units, and a linear layer with
    256 neurons.

    See Also:
        - :class:`Shakespeare_LSTM_E`
        - :class:`Shakespeare_LSTM_D`
    """

    def __init__(self):
        super(Shakespeare_LSTM, self).__init__(Shakespeare_LSTM_E(), Shakespeare_LSTM_D())


class MoonCNN_E(nn.Module):
    """Encoder for the :class:`MoonCNN` network.

    See Also:
        - :class:`MoonCNN`
        - :class:`MoonCNN_D`
    """

    # Expected input size: 32x32x3
    def __init__(self):
        super(MoonCNN_E, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, kernel_size=5)
        self.pool1 = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, kernel_size=5)
        self.pool2 = nn.MaxPool2d(2, 2)
        self.fc1 = nn.Linear(400, 120)
        self.fc2 = nn.Linear(120, 84)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pool1(F.relu(self.conv1(x)))
        x = self.pool2(F.relu(self.conv2(x)))
        x = x.view(-1, 400)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return x


class MoonCNN_D(nn.Module):
    """Head for the :class:`MoonCNN` network.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`MoonCNN`
        - :class:`MoonCNN_E`
    """

    def __init__(self, output_size: int = 10):
        super(MoonCNN_D, self).__init__()
        self.out = nn.Linear(84, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.out(x)
        return x


# MOON: https://arxiv.org/pdf/2103.16257.pdf (CIFAR10)
class MoonCNN(EncoderHeadNet):
    """Convolutional Neural Network for CIFAR-10. This is a CNN for CIFAR-10 classification first
    described in the [MOON]_ paper, where the architecture consists of two convolutional layers with
    6 and 16 filters, respectively, followed by two fully connected layers with 120 and 84 neurons,
    respectively, and a projection head with 256 neurons followed by the output layer with 10
    neurons.

    Args:
        output_size (int, optional): Number of output classes. Defaults to 10.

    See Also:
        - :class:`MoonCNN_E`
        - :class:`MoonCNN_D`

    References:
        .. [MOON] Qinbin Li, Bingsheng He, and Dawn Song. Model-Contrastive Federated Learning.
            In CVPR (2021).
    """

    def __init__(self, output_size: int = 10):
        super(MoonCNN, self).__init__(MoonCNN_E(), MoonCNN_D(output_size))




"""Implementation of the [FedNH23]_ algorithm.

References:
    .. [FedNH23] Yutong Dai, Zeyuan Chen, Junnan Li, Shelby Heinecke, Lichao Sun, Ran Xu.
       Tackling Data Heterogeneity in Federated Learning with Class Prototypes. In AAAI (2023).
       URL: https://arxiv.org/abs/2212.02758
"""
import sys
from collections import defaultdict
from typing import Any, Iterable

import torch
from torch.nn import CrossEntropyLoss, Module, Parameter
from torch.nn import functional as F

sys.path.append(".")
sys.path.append("..")


from .. import FlukeENV  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache, get_model  # NOQA
from ..utils.model import get_activation_size  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "ProtoNet",
    "ArgMaxModule",
    "FedNHClient",
    "FedNHServer",
    "FedNH"
]


class ProtoNet(Module):
    """Neural network with prototypes.
    The network is composed of an encoder and a set of prototypes. The prototypes are
    initialized as orthogonal vectors and are not trainable. In the forward pass, the network
    computes the cosine similarity between the normalized embeddings (output of the encoder) and
    the prototypes.

    Args:
        encoder (Module): Encoder network.
        n_protos (int): Number of prototypes.
        proto_size (int): Size of the prototypes.
    """

    def __init__(self, encoder: Module, n_protos: int, proto_size: int):
        super(ProtoNet, self).__init__()
        self.encoder = encoder
        self.prototypes = Parameter(torch.rand((n_protos, proto_size)), requires_grad=False)
        self.prototypes.data = torch.nn.init.orthogonal_(torch.rand(n_protos, proto_size))
        self.temperature = Parameter(torch.tensor(1.0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        embeddings = self.encoder(x)
        embeddings_norm = torch.norm(embeddings, p=2, dim=1, keepdim=True).clamp(min=1e-12)
        embeddings = torch.div(embeddings, embeddings_norm)
        logits = torch.matmul(embeddings, self.prototypes.T)
        logits = self.temperature * logits
        return embeddings, logits


class ArgMaxModule(Module):
    def __init__(self,
                 model: Module):
        super().__init__()
        self.model: Module = model

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.argmax(self.model(x)[1], dim=1)


class FedNHClient(Client):

    def __init__(self,
                 index: int,
                 model: Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,  # Not used
                 local_epochs: int,
                 n_protos: int,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 5,
                 **kwargs: dict[str, Any]):
        fine_tuning_epochs = fine_tuning_epochs if fine_tuning_epochs > 0 else local_epochs
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=CrossEntropyLoss(),
                         local_epochs=local_epochs, fine_tuning_epochs=fine_tuning_epochs,
                         clipping=clipping, **kwargs)
        self.hyper_params.update(
            n_protos=n_protos
        )
        if isinstance(model, str):
            model = get_model(model)
        embedding_size = get_activation_size(model, train_set.tensors[0][0])
        self.model = ProtoNet(model, n_protos, embedding_size)
        self.count_by_class = torch.bincount(self.train_set.tensors[1],
                                             minlength=self.train_set.num_labels)
        self._save_to_cache()

    def _update_protos(self, protos: Iterable[torch.Tensor]) -> None:
        prototypes = self.model.prototypes.data
        for label, prts in protos.items():
            if prts.shape[0] > 0:
                prototypes[label] = torch.sum(prts, dim=0) / prts.shape[0]
                prototypes[label] /= torch.norm(prototypes[label]).clamp(min=1e-12)
                prototypes[label] = prototypes[label] * prts.shape[0]
            else:
                prototypes[label] = torch.zeros_like(prototypes[label])

    def fit(self, override_local_epochs: int = 0) -> float:

        epochs: int = (override_local_epochs if override_local_epochs
                       else self.hyper_params.local_epochs)

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                _, logits = self.model(X)
                loss = self.hyper_params.loss_fn(logits, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()
        self.model.cpu()
        clear_cuda_cache()
        running_loss /= (epochs * len(self.train_set))

        protos = defaultdict(list)
        for label in range(self.hyper_params.n_protos):
            Xlbl = self.train_set.tensors[0][self.train_set.tensors[1] == label]
            protos[label] = self.model.encoder(Xlbl).detach().data

        self._update_protos(protos)
        return running_loss

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if test_set is not None and self.model is not None:
            model = ArgMaxModule(self.model)
            return evaluator.evaluate(self._last_round, model, test_set, device=self.device)
        return {}

    def finalize(self) -> None:
        self._load_from_cache()
        self.fit(self.hyper_params.fine_tuning_epochs)
        metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
        if metrics:
            self._notify_evaluation(-1, "post-fit", metrics)
        self._save_to_cache()


class FedNHServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 n_protos: int = 10,
                 rho: float = 0.1):
        embedding_size = get_activation_size(model, clients[0].train_set.tensors[0][0])
        super().__init__(model=ProtoNet(model, n_protos, embedding_size),
                         test_set=test_set,
                         clients=clients,
                         weighted=weighted)
        self.hyper_params.update(n_protos=n_protos, rho=rho)

    @torch.no_grad()
    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:

        # This could be the learning rate for the server (not used in the official implementation)
        # server_lr = self.hyper_params.lr * self.hyper_params.lr_decay ** self.round
        server_lr = 1.0
        weights = [server_lr / len(eligible)] * len(eligible)
        cl_weight = torch.zeros(self.hyper_params.n_protos)

        # To get client.count_by_class is actually illegal in fluke, but irrelevant from an
        # implementation point of view
        for client in eligible:
            cl_weight += client.count_by_class

        clients_protos = []

        # Aggregate models = Federated Averaging
        # Get model parameters and buffers
        model_params = dict(self.model.encoder.named_parameters())
        # Includes running_mean, running_var, etc.
        model_buffers = dict(self.model.encoder.named_buffers())

        # Initialize accumulators for parameters
        avg_params = {key: torch.zeros_like(param.data) for key, param in model_params.items()}
        avg_buffers = {key: torch.zeros_like(buffer.data)
                       for key, buffer in model_buffers.items() if "num_batches_tracked" not in key}

        max_num_batches_tracked = 0  # Track the max num_batches_tracked

        # Compute weighted sum (weights already sum to 1, so no division needed)
        for m, w in zip(client_models, weights):
            clients_protos.append(m.prototypes.data)

            for key, param in m.encoder.named_parameters():
                avg_params[key].add_(param.data, alpha=w)

            for key, buffer in m.encoder.named_buffers():
                if "num_batches_tracked" not in key:
                    avg_buffers[key].add_(buffer.data, alpha=w)
                else:
                    max_num_batches_tracked = max(max_num_batches_tracked, buffer.item())

        for key in model_params.keys():
            model_params[key].data.lerp_(avg_params[key], server_lr)  # Soft update

        for key in model_buffers.keys():
            if "num_batches_tracked" not in key:
                model_buffers[key].data.lerp_(avg_buffers[key], server_lr)

        # Assign max num_batches_tracked
        for key in model_buffers.keys():
            if "num_batches_tracked" in key:
                model_buffers[key].data.fill_(max_num_batches_tracked)

        # Group by label
        label_protos = {i: [protos[i] for protos in clients_protos]
                        for i in range(self.hyper_params.n_protos)}

        # Aggregate prototypes
        prototypes = self.model.prototypes.clone()
        if self.hyper_params.weighted:
            for label, protos in label_protos.items():
                prototypes.data[label, :] = torch.sum(torch.stack(protos) / cl_weight[label], dim=0)
        else:
            sim_weights = []
            for protos in clients_protos:
                sim_weights.append(torch.exp(torch.sum(prototypes.data * protos, dim=1)))
            sim_weights = torch.stack(sim_weights, dim=0).T

            for label, protos in label_protos.items():
                prototypes.data[label, :] = prototypes.data[label, :] + \
                    torch.sum(sim_weights[label].unsqueeze(1) * torch.stack(protos), dim=0)

            prototypes.data /= torch.sum(sim_weights, dim=1).unsqueeze(1)

        # Normalize the prototypes
        # prototypes.data /= torch.norm(prototypes.data, dim=0).clamp(min=1e-12)
        prototypes.data = F.normalize(prototypes.data, dim=1)  # .clamp(min=1e-12)

        self.model.prototypes.data = (1 - self.hyper_params.rho) * prototypes.data + \
            self.hyper_params.rho * self.model.prototypes.data

        # Normalize the prototypes again
        self.model.prototypes.data = F.normalize(
            self.model.prototypes.data, dim=1)  # .clamp(min=1e-12)

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if self.test_set is not None:
            model = ArgMaxModule(self.model)
            return evaluator.evaluate(self.rounds + 1, model, self.test_set, device=self.device)
        return {}


class FedNH(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedNHClient

    def get_server_class(self) -> Server:
        return FedNHServer



"""This module contains utility classes for loading datasets."""
import os
import sys
from typing import Any

import torch
from PIL import Image
from rich.progress import track
from torchvision.datasets import ImageFolder, VisionDataset
from torchvision.datasets.utils import download_and_extract_archive
from torchvision.transforms import ToTensor

sys.path.append(".")
sys.path.append("..")

__all__ = [
    "MNISTM",
    "CINIC10"
]


class MNISTM(VisionDataset):
    """MNIST-M Dataset.
    """

    resources = [
        ('https://github.com/liyxi/mnist-m/releases/download/data/mnist_m_train.pt.tar.gz',
         '191ed53db9933bd85cc9700558847391'),
        ('https://github.com/liyxi/mnist-m/releases/download/data/mnist_m_test.pt.tar.gz',
         'e11cb4d7fff76d7ec588b1134907db59')
    ]

    training_file = "mnist_m_train.pt"
    test_file = "mnist_m_test.pt"
    classes = ['0 - zero', '1 - one', '2 - two', '3 - three', '4 - four',
               '5 - five', '6 - six', '7 - seven', '8 - eight', '9 - nine']

    def __init__(self, root, train=True, transform=None, target_transform=None, download=True):
        """Init MNIST-M dataset."""
        super(MNISTM, self).__init__(root, transform=transform, target_transform=target_transform)

        self.train = train

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError("Dataset not found." +
                               " You can use download=True to download it")

        data_file = (self.training_file if self.train
                     else self.test_file)
        self.data, self.targets = torch.load(os.path.join(self.processed_folder, data_file),
                                             weights_only=False)

    def __getitem__(self, index):
        """Get images and target for data loader.

        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], int(self.targets[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.squeeze().numpy(), mode="RGB")

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        """Return size of dataset."""
        return len(self.data)

    @property
    def raw_folder(self):
        return os.path.join(self.root, self.__class__.__name__, 'raw')

    @property
    def processed_folder(self):
        return os.path.join(self.root, self.__class__.__name__, 'processed')

    @property
    def class_to_idx(self):
        return {_class: i for i, _class in enumerate(self.classes)}

    def _check_exists(self):
        return (os.path.exists(os.path.join(self.processed_folder, self.training_file)) and
                os.path.exists(os.path.join(self.processed_folder, self.test_file)))

    def download(self):
        """Download the MNIST-M data."""

        if self._check_exists():
            return

        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)

        # download files
        for url, md5 in self.resources:
            filename = url.rpartition('/')[2]
            download_and_extract_archive(url, download_root=self.raw_folder,
                                         extract_root=self.processed_folder,
                                         filename=filename, md5=md5)


class CINIC10(VisionDataset):
    """`CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset. CINIC-10 is an augmented
    extension of CIFAR-10. It contains the images from CIFAR-10 (60,000 images, 32x32 RGB pixels)
    and a selection of ImageNet database images (210,000 images downsampled to 32x32). It was
    compiled as a 'bridge' between CIFAR-10 and ImageNet, for benchmarking machine learning
    applications. It is split into three equal subsets - train, validation, and test -
    each of which contain 90,000 images.

    Args:
        root (string): Root directory of dataset where directory ``cifar-10-batches-py`` exists
            or will be saved to if download is set to True.
        partition (str, optional): One of train,valid,test, creates selects which partition to
            use.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If true, downloads the dataset from the internet and puts it
            in root directory. If dataset is already downloaded, it is not downloaded again.
    """

    base_folder = "CINIC10"
    filename = "CINIC-10.tar.gz"
    url = "https://datashare.is.ed.ac.uk/bitstream/handle/10283/3192/CINIC-10.tar.gz"
    file_md5 = "6ee4d0c996905fe93221de577967a372"
    # mean = [0.47889522, 0.47227842, 0.43047404]
    # std = [0.24205776, 0.23828046, 0.25874835]

    classes = [
        "airplane",
        "automobile",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck"
    ]

    def __init__(self,
                 root: str = "../data",
                 split: str = "train",
                 #  transform: Optional[Callable] = None,
                 download: bool = True):

        super().__init__(root)
        self.split = split
        self.root = os.path.join(root, "CINIC10")

        if download:
            self.download()

        self._dataset = ImageFolder(os.path.join(self.root, split))
        self.data = self._images2tensor()
        self.targets = torch.LongTensor(self._dataset.targets)

    def _images2tensor(self):
        img_tensors = []
        for img, _ in track(self._dataset.imgs, f"Loading CINIC-10 {self.split} set..."):
            img_tensor = ToTensor()(Image.open(img))
            if img_tensor.shape[0] == 1:
                img_tensor = img_tensor.repeat(3, 1, 1)
            img_tensors.append(img_tensor)

        return torch.stack(img_tensors)

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        return self.data[index], self.targets[index]

    def __len__(self) -> int:
        return len(self.data)

    def download(self) -> None:
        os.makedirs(self.root, exist_ok=True)
        if os.path.isdir(os.path.join(self.root, self.split)):
            return
        else:
            download_and_extract_archive(self.url,
                                         download_root=self.root,
                                         md5=CINIC10.file_md5)

    @property
    def class_to_idx(self):
        return {c: i for i, c in enumerate(CINIC10.classes)}



"""Implementation of the [FedProx18]_ algorithm.

References:
    .. [FedProx18] Tian Li, Anit Kumar Sahu, Manzil Zaheer, Maziar Sanjabi, Ameet Talwalkar,
       and Virginia Smith. Federated Optimization in Heterogeneous Networks. Adaptive & Multitask
       Learning Workshop. URL: https://openreview.net/pdf?id=SkgwE5Ss3N
"""
import sys
from copy import deepcopy
from typing import Any

import torch

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from . import CentralizedFL  # NOQA


__all__ = [
    "FedProxClient",
    "FedProx"
]


class FedProxClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 mu: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(mu=mu)

    def _proximal_loss(self, local_model, global_model):
        proximal_term = 0.0
        for w, w_t in zip(local_model.parameters(), global_model.parameters()):
            if w_t.requires_grad:
                proximal_term += (w - w_t).norm(2) ** 2
        return proximal_term

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        W = deepcopy(self.model).to(self.device)
        self.model.to(self.device)
        self.model.train()

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y) + \
                    (self.hyper_params.mu / 2) * self._proximal_loss(self.model, W)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        W.cpu()
        clear_cuda_cache()
        return running_loss


class FedProx(CentralizedFL):

    def get_client_class(self) -> Client:
        return FedProxClient



"""Implementation of the DITTO [Ditto21]_ algorithm.

References:
    .. [Ditto21] Tian Li, Shengyuan Hu, Ahmad Beirami, and Virginia Smith. Ditto: Fair and Robust
       Federated Learning Through Personalization. In ICML (2021).
       URL: https://arxiv.org/abs/2012.04221
"""
import sys
from copy import deepcopy
from typing import Any, Iterator

import torch
from torch.nn.parameter import Parameter
from torch.optim import Optimizer

sys.path.append(".")
sys.path.append("..")

from ..client import PFLClient  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from . import PersonalizedFL  # NOQA

__all__ = [
    "PerturbedGradientDescent",
    "DittoClient",
    "Ditto"
]


class PerturbedGradientDescent(Optimizer):
    def __init__(self,
                 params: Iterator[Parameter],
                 lr: float = 0.01,
                 lam: float = 0.0,
                 **kwargs: dict[str, Any]):
        default = dict(lr=lr, lam=lam)
        super().__init__(params, default)

    @torch.no_grad()
    def step(self, global_params):
        for group in self.param_groups:
            for p, g in zip(group['params'], global_params):
                d_p = p.grad.data + group['lam'] * (p.data - g.data)
                p.data.add_(d_p, alpha=-group['lr'])


# https://arxiv.org/pdf/2012.04221.pdf
class DittoClient(PFLClient):

    def __init__(self,
                 index: int,
                 model: torch.nn.Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 tau: int = 3,
                 lam: float = 0.1,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, model=model, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.pers_optimizer = None
        self.pers_scheduler = None
        self.hyper_params.update(
            tau=tau,
            lam=lam
        )
        self._save_to_cache()

    def _proximal_loss(self, local_model, global_model):
        proximal_term = 0.0
        for name, param in local_model.named_parameters():
            # if 'weight' not in name:
            #     continue
            proximal_term += (param - global_model.get_parameter(name))  # .norm(2)
        return proximal_term

    def fit(self, override_local_epochs: int = 0) -> dict:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)

        w_prev = deepcopy(self.model)

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
            self.scheduler.step()

        self.model.cpu()
        clear_cuda_cache()

        self.personalized_model.train()
        self.personalized_model.to(self.device)
        w_prev.to(self.device)

        if self.pers_optimizer is None:
            self.pers_optimizer = PerturbedGradientDescent(self.personalized_model.parameters(),
                                                           lam=self.hyper_params.lam,
                                                           **self._optimizer_cfg.optimizer_cfg)
            self.pers_scheduler = self._optimizer_cfg.scheduler(
                self.pers_optimizer,
                **self._optimizer_cfg.scheduler_cfg
            )

        running_loss = 0.0
        for _ in range(self.hyper_params.tau):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.pers_optimizer.zero_grad()
                y_hat = self.personalized_model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self.pers_optimizer.step(w_prev.parameters())
                running_loss += loss.item()
            self.pers_scheduler.step()

        running_loss /= (self.hyper_params.tau * len(self.train_set))
        self.personalized_model.cpu()
        w_prev.cpu()
        clear_cuda_cache()

        return running_loss


class Ditto(PersonalizedFL):

    def get_client_class(self) -> PFLClient:
        return DittoClient



"""Implementation of the [Moon21]_ algorithm.

References:
    .. [Moon21] Qinbin Li, Bingsheng He, and Dawn Song. Model-Contrastive Federated Learning.
       In CVPR (2021). URL: https://arxiv.org/abs/2103.16257
"""
import sys
from copy import deepcopy
from typing import Any

import torch
from torch.nn import CosineSimilarity

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA

__all__ = [
    "MOONClient",
    "MOON"
]


class MOONClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 mu: float,
                 tau: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(
            mu=mu,
            tau=tau
        )
        self.prev_model = None
        self.server_model = None
        self._attr_to_cache.extend(["prev_model", "server_model"])

    def receive_model(self) -> None:
        model = self.channel.receive(self, self.server, msg_type="model").payload
        if self.model is None:
            # one of these deepcopy is not needed
            self.model = deepcopy(model)
            self.prev_model = deepcopy(model)
        else:
            self.prev_model.load_state_dict(deepcopy(self.model.state_dict()))
            safe_load_state_dict(self.model, deepcopy(model.state_dict()))
        self.server_model = model

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        cos = CosineSimilarity(dim=-1).to(self.device)
        self.model.to(self.device)
        self.prev_model.to(self.device)
        self.server_model.to(self.device)
        self.model.train()
        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()

                z_local = self.model.encoder(X)  # , -1)
                y_hat = self.model.head(z_local)
                loss_sup = self.hyper_params.loss_fn(y_hat, y)

                z_prev = self.prev_model.encoder(X)  # , -1)
                z_global = self.server_model.encoder(X)  # , -1)

                sim_lg = cos(z_local, z_global).reshape(-1, 1) / self.hyper_params.tau
                sim_lp = cos(z_local, z_prev).reshape(-1, 1) / self.hyper_params.tau
                loss_con = -torch.log(torch.exp(sim_lg) /
                                      (torch.exp(sim_lg) + torch.exp(sim_lp))).mean()

                loss = loss_sup + self.hyper_params.mu * loss_con
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= (epochs * len(self.train_set))
        self.prev_model.cpu()
        self.server_model.cpu()
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class MOON(CentralizedFL):

    def get_client_class(self) -> Client:
        return MOONClient



"""Implementation of the DPFedAVG: Differential Privacy Federated Averaging [DPFedAVG2017]_
algorithm.

Note:
    This implementation does not exactly follow the original paper, but it is a simplified version
    that uses the Opacus library to provide differential privacy guarantees.

References:
    .. [DPFedAVG2017] Robin C. Geyer, Tassilo Klein, Moin Nabi.
       Differentially Private Federated Learning: A Client Level Perspective
       In ArXiv (2017). URL: https://arxiv.org/abs/1712.07557
"""
import sys
from typing import Iterable, Any

import torch
from opacus import PrivacyEngine
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator  # NOQA
from . import CentralizedFL  # NOQA

__all__ = [
    "DPFedAVG",
    "DPFedAVGClient",
    "DPFedAVGServer"
]


class _OpacusModelAdapter(Module):
    """Adapt a model to be compatible with Opacus.
    Opacus encapsultes the model in a way that the parameter names have a prefix
    named "_module". This class is a simple adapter to make the model compatible
    with Opacus.

    Args:
        model (Module): The model to be adapted.
    """

    def __init__(self, model: Module):
        super().__init__()
        self._module = model

    def forward(self, *args, **kwargs) -> torch.Tensor:
        return self._module(*args, **kwargs)


class DPFedAVGClient(Client):
    def __init__(self,
                 index: int,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: Module,
                 local_epochs: int = 3,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 noise_mul: float = 1.1,
                 max_grad_norm: float = 1.0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index, train_set=train_set, test_set=test_set,
                         optimizer_cfg=optimizer_cfg, loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        self.hyper_params.update(noise_mul=noise_mul, max_grad_norm=max_grad_norm)

    def _init_private_engine(self) -> None:
        if self.model is None:
            return
        self.privacy_engine = PrivacyEngine()
        self.model.train()
        self.model, self.optimizer, self.train_set = self.privacy_engine.make_private(
            module=self.model._module,
            optimizer=self.optimizer,
            data_loader=self.train_set.asDataLoader() if isinstance(
                self.train_set, FastDataLoader) else self.train_set,
            noise_multiplier=self.hyper_params.noise_mul,
            max_grad_norm=self.hyper_params.max_grad_norm,

        )

    def receive_model(self) -> None:
        if self.model is None:
            super().receive_model()
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)
            self._init_private_engine()
        else:
            return super().receive_model()


class DPFedAVGServer(Server):

    def __init__(self,
                 model: Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 lr: float = 1.0,
                 **kwargs: dict[str, Any]):
        super().__init__(model=_OpacusModelAdapter(model),
                         test_set=test_set, clients=clients, weighted=weighted, lr=lr, **kwargs)


class DPFedAVG(CentralizedFL):

    def get_server_class(self) -> Server:
        return DPFedAVGServer

    def get_client_class(self) -> Client:
        return DPFedAVGClient



"""Implementation of the FedHP: Federated Learning with Hyperspherical Prototypical Regularization
[FedHP24]_ algorithm.

References:
    .. [FedHP24] Samuele Fonio, Mirko Polato, Roberto Esposito.
       FedHP: Federated Learning with Hyperspherical Prototypical Regularization. In ESANN (2024).
       URL: https://www.esann.org/sites/default/files/proceedings/2024/ES2024-183.pdf
"""
import copy
import sys
from typing import Any, Iterable, Literal, Generator

import torch
import torch.optim as optim
from rich.progress import track
from torch import nn
from torch.optim.optimizer import Optimizer as Optimizer

sys.path.append(".")
sys.path.append("..")
from .. import FlukeENV  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..data import FastDataLoader  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..server import Server  # NOQA
from ..utils import OptimizerConfigurator, clear_cuda_cache  # NOQA
from ..utils.model import get_activation_size  # NOQA
from . import PersonalizedFL  # NOQA


class ProtoNet(nn.Module):
    def __init__(self, encoder: nn.Module, n_protos: int, proto_size: int):
        super(ProtoNet, self).__init__()
        self._encoder = encoder
        self.prototypes = nn.Parameter(torch.rand((n_protos, proto_size)),
                                       requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        embeddings = self._encoder(x)
        dists = -torch.norm(embeddings[:, None, :] - self.prototypes[None, :, :], dim=-1)
        return embeddings, dists


class FedHPModel(nn.Module):
    def __init__(self,
                 model: nn.Module):
        super().__init__()
        self.model: nn.Module = model

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.argmax(self.model(x)[1], dim=1)


class SeparationLoss(nn.Module):
    """Large margin separation between hyperspherical protoypes"""

    def __init__(self, reduction: Literal["mean", "sum"] = "mean"):
        super(SeparationLoss, self).__init__()
        self.reduction = reduction

    def forward(self, protos: torch.Tensor) -> torch.Tensor:
        # protos: (N_prototypes x Embedding_dimension)
        M = torch.matmul(protos, protos.transpose(0, 1)) - 2 * torch.eye(
            protos.shape[0]).to(protos.device)
        loss = M.max(dim=1)[0]
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        return loss


class FedHPClient(Client):
    def __init__(self,
                 index: int,
                 model: nn.Module,
                 train_set: FastDataLoader,
                 test_set: FastDataLoader,
                 optimizer_cfg: OptimizerConfigurator,
                 loss_fn: torch.nn.Module,
                 local_epochs: int,
                 n_protos: int,
                 lam: float,
                 fine_tuning_epochs: int = 0,
                 clipping: float = 0,
                 **kwargs: dict[str, Any]):
        super().__init__(index=index,
                         train_set=train_set, test_set=test_set, optimizer_cfg=optimizer_cfg,
                         loss_fn=loss_fn, local_epochs=local_epochs,
                         fine_tuning_epochs=fine_tuning_epochs, clipping=clipping, **kwargs)
        embedding_size = get_activation_size(model, train_set.tensors[0][0])
        self.model = ProtoNet(model, n_protos, embedding_size)
        self.hyper_params.update(n_protos=n_protos, lam=lam)
        self.anchors = None
        # This optimizer must be saved in the cache along with the model from which its parameters
        # are taken. See the property `proto_opt`.
        self._modopt.additional = {"proto_opt": None}
        self._attr_to_cache.append("anchors")
        self._save_to_cache()

    @property
    def proto_opt(self) -> Optimizer:
        """The optimizer for the prototypes.

        Returns:
            Optimizer: The optimizer for the prototypes.
        """
        return self._modopt.additional["proto_opt"]

    @proto_opt.setter
    def proto_opt(self, value: Optimizer) -> None:
        self._modopt.additional["proto_opt"] = value

    def receive_model(self) -> None:
        if self.anchors is None:
            msg = self.channel.receive(self, self.server, msg_type="anchors")
            self.anchors = msg.payload.data
        msg = self.channel.receive(self, self.server, msg_type="prototypes")
        self.model.prototypes.data = msg.payload

    def send_model(self) -> None:
        self.channel.send(Message(copy.deepcopy(self.model.prototypes.data),
                          "prototypes", self, inmemory=True), self.server)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (override_local_epochs if override_local_epochs > 0
                       else self.hyper_params.local_epochs)
        self.model.train()
        self.model.to(self.device)

        def filter_fun(model): return [param for name, param in model.named_parameters()
                                       if 'prototype' not in name]

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(
                self.model, filter_fun=filter_fun)

        if self.proto_opt is None:
            proto_params = [p for name, p in self.model.named_parameters()
                            if 'proto' in name]
            self.proto_opt = optim.Adam(proto_params, lr=0.005)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                self.proto_opt.zero_grad()
                _, dists = self.model.forward(X)
                loss = self.hyper_params.loss_fn(dists, y)
                loss_proto = torch.mean(1 - nn.CosineSimilarity(dim=1)
                                        (self.model.prototypes, self.anchors))
                loss += self.hyper_params.lam * loss_proto
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                self.proto_opt.step()
                running_loss += loss.item()
            self.scheduler.step()
        running_loss /= (epochs * len(self.train_set))
        self.model.cpu()
        clear_cuda_cache()
        return running_loss

    def evaluate(self, evaluator: Evaluator, test_set: FastDataLoader) -> dict[str, float]:
        if test_set is not None and self.anchors is not None:
            model = FedHPModel(self.model)
            return evaluator.evaluate(self._last_round, model, test_set, device=self.device)
        return {}

    def finalize(self) -> None:
        self._load_from_cache()
        self.receive_model()
        self.fit(self.hyper_params.fine_tuning_epochs)
        metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
        if metrics:
            self._notify_evaluation(-1, "post-fit", metrics)
        self._save_to_cache()


class FedHPServer(Server):
    def __init__(self,
                 model: nn.Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = True,
                 n_protos: int = 10,
                 embedding_size: int = 100,
                 **kwargs: dict[str, Any]):
        super().__init__(model=ProtoNet(model, n_protos, embedding_size),
                         test_set=None,
                         clients=clients,
                         weighted=weighted)
        self.hyper_params.update(n_protos=n_protos,
                                 embedding_size=embedding_size)
        self.device = FlukeENV().get_device()
        self.anchors = None
        self.prototypes = None
        self.clients_class_weights = None

    def fit(self, n_rounds: int = 10, eligible_perc: float = 0.1, finalize: bool = True) -> None:
        if self.rounds == 0:
            self.anchors = self._hyperspherical_embedding().data
            self.channel.broadcast(Message(self.anchors, "anchors", self), self.clients)
            self.prototypes = copy.deepcopy(self.anchors)
            client = {c.index: c.train_set.tensors[1] for c in self.clients}
            # Count the occurrences of each class for each client.
            # This is "illegal" in fluke :)
            n_classes = self.clients[0].train_set.num_labels
            class_counts = {client_idx: torch.bincount(client_data, minlength=n_classes).tolist()
                            for client_idx, client_data in enumerate(client.values())}
            if self.hyper_params.weighted:
                tensor_class_counts = torch.empty((len(class_counts[0]), self.n_clients))
                for ind, val in enumerate(class_counts.values()):
                    tensor_class_counts[:, ind] = torch.tensor(val)
                col_sums = tensor_class_counts.sum(dim=0, keepdim=True)
                self.clients_class_weights = tensor_class_counts / col_sums
            else:
                self.clients_class_weights = torch.ones((len(class_counts[0]), self.n_clients))

        return super().fit(n_rounds=n_rounds, eligible_perc=eligible_perc, finalize=finalize)

    def _hyperspherical_embedding(self):
        """
        Function to learn the prototypes according to the ``SeparationLoss`` minimization.
        """
        lr = 0.1
        momentum = 0.9
        n_steps = 1000
        wd = 1e-4
        # torch.manual_seed(seed)
        mapping = torch.rand((self.hyper_params.n_protos, self.hyper_params.embedding_size),
                             device=self.device, requires_grad=True)
        optimizer = torch.optim.SGD([mapping], lr=lr, momentum=momentum, weight_decay=wd)
        loss_fn = SeparationLoss()
        for _ in track(range(n_steps), "[SERVER] Learning prototypes..."):
            with torch.no_grad():
                mapping.div_(torch.norm(mapping, dim=1, keepdim=True))
            optimizer.zero_grad()
            loss = loss_fn(mapping)
            loss.backward()
            optimizer.step()
        with torch.no_grad():
            mapping.div_(torch.norm(mapping, dim=1, keepdim=True))
        return mapping.detach()

    def broadcast_model(self, eligible: Iterable[FedHPClient]) -> None:
        self.channel.broadcast(Message(self.prototypes, "prototypes", self), eligible)

    def receive_client_models(self,
                              eligible: Iterable[Client],
                              state_dict: bool = False) -> Generator[nn.Module, None, None]:
        for client in eligible:
            yield self.channel.receive(self, client, "prototypes").payload

    def aggregate(self,
                  eligible: Iterable[FedHPClient],
                  client_models: Iterable[nn.Module]) -> None:
        clients_prototypes = client_models
        clients_weights = self.clients_class_weights[:, [client.index for client in eligible]].T
        avg_proto = None

        for i, client_proto in enumerate(clients_prototypes):
            if avg_proto is None:
                avg_proto = torch.zeros_like(client_proto)
            avg_proto += client_proto * clients_weights[i, :].unsqueeze(-1)
        avg_proto /= clients_weights.sum(dim=0).unsqueeze(-1)
        self.prototypes = avg_proto


class FedHP(PersonalizedFL):

    def get_client_class(self) -> Client:
        return FedHPClient

    def get_server_class(self) -> Server:
        return FedHPServer

    def get_optimizer_class(self) -> Optimizer:
        return torch.optim.Adam



"""Implementation of the Kafè [Kafe24]_ algorithm.

References:
    .. [Kafe24] Pian Qi, Diletta Chiaro, Fabio Giampaolo, and Francesco Piccialli.
       KAFÈ: Kernel Aggregation for FEderated. In ECML-PKDD (2024).
       URL: https://link.springer.com/content/pdf/10.1007/978-3-031-70359-1_4.pdf

"""
import sys
from typing import Iterable

import torch
from torch.nn import Module
import numpy as np
from sklearn.neighbors import KernelDensity

sys.path.append(".")
sys.path.append("..")

from . import CentralizedFL  # NOQA
from ..data import FastDataLoader  # NOQA
from ..client import Client  # NOQA
from ..server import Server  # NOQA

__all__ = [
    "KafeServer",
    "Kafe"
]


class KafeServer(Server):

    def __init__(self,
                 model: torch.nn.Module,
                 test_set: FastDataLoader,
                 clients: Iterable[Client],
                 weighted: bool = False,
                 bandwidth: float = 1.0):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(bandwidth=bandwidth)

    def aggregate(self, eligible: Iterable[Client], client_models: Iterable[Module]) -> None:
        weights = self._get_client_weights(eligible)

        # get last layer of m clients' weights
        last_layer_weight_name = list(self.model.state_dict().keys())[-2]
        last_layer_bias_name = list(self.model.state_dict().keys())[-1]

        # Get model parameters and buffers
        model_params = dict(self.model.named_parameters())
        model_buffers = dict(self.model.named_buffers())  # Includes running_mean, running_var, etc.

        # Initialize accumulators for parameters
        avg_params = {key: torch.zeros_like(param.data) for key, param in model_params.items()}
        avg_buffers = {key: torch.zeros_like(buffer.data)
                       for key, buffer in model_buffers.items() if "num_batches_tracked" not in key}

        max_num_batches_tracked = 0  # Track the max num_batches_tracked
        w_last_layer = []
        b_last_layer = []

        # Compute weighted sum (weights already sum to 1, so no division needed)
        for m, w in zip(client_models, weights):
            for key, param in m.named_parameters():
                if key == last_layer_weight_name:
                    w_last_layer.append(param.data.numpy().copy())
                    continue

                if key == last_layer_bias_name:
                    b_last_layer.append(param.data.numpy().copy())
                    continue

                avg_params[key].add_(param.data, alpha=w)

            for key, buffer in m.named_buffers():
                if "num_batches_tracked" not in key:
                    avg_buffers[key].add_(buffer.data, alpha=w)
                else:
                    max_num_batches_tracked = max(max_num_batches_tracked, buffer.item())

        for key in model_params.keys():
            if key in (last_layer_weight_name, last_layer_bias_name):
                continue
            model_params[key].data.lerp_(avg_params[key], self.hyper_params.lr)  # Soft update

        for key in model_buffers.keys():
            if "num_batches_tracked" not in key:
                model_buffers[key].data.lerp_(avg_buffers[key], self.hyper_params.lr)  # Soft update

        # Assign max num_batches_tracked
        for key in model_buffers.keys():
            if "num_batches_tracked" in key:
                model_buffers[key].data.fill_(max_num_batches_tracked)

        w_last_layer = np.array(w_last_layer).reshape(len(w_last_layer), -1)
        b_last_layer = np.array(b_last_layer).reshape(len(b_last_layer), -1)

        # using KDE get the kernel density of last layers
        kde_w = KernelDensity(kernel='gaussian',
                              bandwidth=self.hyper_params.bandwidth).fit(w_last_layer,
                                                                         sample_weight=weights)
        kde_b = KernelDensity(kernel='gaussian',
                              bandwidth=self.hyper_params.bandwidth).fit(b_last_layer,
                                                                         sample_weight=weights)

        # sample m samples and average, then obtain a new last layer for the global model
        w_last_layer_new = np.mean(kde_w.sample(len(w_last_layer)), axis=0)
        b_last_layer_new = np.mean(kde_b.sample(len(b_last_layer)), axis=0)

        model_sd = self.model.state_dict()
        model_sd[last_layer_weight_name].data = torch.tensor(w_last_layer_new.reshape(
            model_sd[last_layer_weight_name].shape))
        model_sd[last_layer_bias_name].data = torch.tensor(b_last_layer_new.reshape(
            model_sd[last_layer_bias_name].shape))


class Kafe(CentralizedFL):

    def get_server_class(self) -> Server:
        return KafeServer
